/* -*- Mode: C; c-file-style: "stroustrup"; encoding: latin-1-unix -*- */
/* NxpNameSpaces + NxpDefinitions */

#ifndef __NXP_NAMESPACE_H__
#define __NXP_NAMESPACE_H__

/* --- defines --- */

#define NXP_DEF_VARIABLE(def) 	((NxpDefVariable *)def)
#define NXP_DEF_BI_FUNC(def) 	((NxpDefBIFunc   *)def)
#define NXP_DEF_USER_FUNC(def) 	((NxpDefUserFunc *)def)
#define NXP_DEFINITION(def) 	((NxpDefinition  *)def)

#define nxp_definition_destroy nxp_definition_unref
#define nxp_name_space_get_default() (nxp_element_context_get()->name_space)



/*  --===-- forward declarations  --===-- */
typedef struct _NxpDefinition 	NxpDefinition;
typedef struct _NxpDefVariable 	NxpDefVariable;
typedef struct _NxpDefBIFunc 	NxpDefBIFunc;
typedef struct _NxpDefUserFunc 	NxpDefUserFunc;
typedef struct _NxpNameSpace 	NxpNameSpace;





/* --- enums  --- */

typedef enum {
    NXP_DEF_VARIABLE = 1,
    NXP_DEF_BI_FUNC,
    NXP_DEF_USER_FUNC
} NxpDefinitionType;





/* --- includes  --- */
#include <glib.h>
#include "element.h"
#include "module.h"
#include "identifier.h"
#include "nxp_config.h"

/* --- structures  --- */

/* Base class for all definitions */
struct _NxpDefinition
{
    NxpDefinitionType type;
    volatile guint ref_count;
    char *name;
    gboolean read_only:1;
    NxpModule *module; 		/* can be NULL (for user defined variables, for example) */
    
#ifdef NXP_NAMESPACE_DEBUG
    NxpNameSpace *parent_namespace;
#endif
};


/* A variable: contains an element */
struct _NxpDefVariable 
{
    NxpDefinition definition;
    NxpElement *value;
};



/* A user-defined function. Just like a variable, but with a list of
   strings for parameters. The strings are owned by the definition. */
struct _NxpDefUserFunc 
{
    NxpDefinition definition;
    /* arguments: a list of strings (char *)  */
    GSList *argument_list;
    /* A string chunk that holds all the strings. */ 
    GStringChunk *argument_strings; 
    /* The function body */
    NxpElement *expression;
};


typedef NxpElement * (*NxpBIFuncEval)		(GSList *args);
typedef NxpElement * (*NxpBIFuncDerive)		(NxpIdentifier *var,
						 GSList *args);
typedef NxpElement * (*NxpBIFuncSimplify)	(GSList *args);


/* A built-in function.  */
struct _NxpDefBIFunc 
{
    NxpDefinition definition;
    GSList *prototypes; /* data is pointer to NxpProtoInstance */
};

/* --- Listeners --- */
typedef void (*NxpNSListenerFunc) (NxpNameSpace *ns, NxpEvalContext *context, gpointer data);

typedef struct _NxpNSListener NxpNSListener;

struct _NxpNSListener {
    NxpNSListenerFunc  func;
    GSList            *nslist;	/* list of namespaces that have changed */
    NxpEvalContext    *context;
    gpointer           user_data;
};


struct _NxpNameSpace {
    GHashTable      *definitions;
    NxpNameSpace    *parent;
    NxpNameSpace    *next;
    NxpNameSpace    *child;
    pthread_mutex_t  mutex;
    gboolean         reentrant:1;
    gchar           *name;
    GSList          *listeners;	/* list of NxpNSListener */
    volatile gint    modified;

    NxpModule *module; 		/* can be NULL (for user defined variables, for example) */
};







/* definition functions */
NxpDefinition*  nxp_definition_ref                (NxpDefinition     *def);
void            nxp_definition_unref              (NxpDefinition     *def);
NxpDefVariable* nxp_def_variable_new              (char              *name,
						   NxpElement        *value);
NxpDefBIFunc*   nxp_def_bi_func_new               (char              *name);
NxpDefUserFunc* nxp_def_user_func_new             (char              *name,
						   NxpElement        *expression);
void            nxp_def_user_func_append_argument (NxpDefUserFunc    *def,
						   const char        *arg);


/* name space functions */
GString*       nxp_name_space_get_full_path_name         (NxpNameSpace   *self);
NxpNameSpace*  nxp_name_space_new                        (NxpNameSpace   *parent,
							  const char     *name);
void           nxp_name_space_add_definition             (NxpNameSpace   *self,
							  NxpDefinition  *def);
NxpDefinition* nxp_name_space_lookup_definition          (NxpNameSpace   *self,
							  const char     *name);
NxpDefinition* nxp_name_space_lookup_definition_extended (NxpNameSpace   *self,
							  const char     *name,
							  NxpNameSpace  **actual_name_space,
							  gboolean        recursive);
void           nxp_name_space_remove_definition          (NxpNameSpace   *self,
							  NxpDefinition  *def);
void           nxp_name_space_destroy                    (NxpNameSpace   *self);
void           nxp_name_space_lock                       (NxpNameSpace   *self);
void           nxp_name_space_unlock                     (NxpNameSpace   *self);
NxpNameSpace*  nxp_name_space_find_child                 (NxpNameSpace   *self,
							  const char     *name);
gboolean       nxp_name_space_resolve_path               (NxpNameSpace   *cwd,
							  NxpIdentifier  *ident);



/* reentrant versions of some functions */
void           nxp_name_space_add_definition_r             (NxpNameSpace   *self,
							    NxpDefinition  *def);
void           nxp_name_space_remove_definition_r          (NxpNameSpace   *self,
							    NxpDefinition  *def);
NxpDefinition* nxp_name_space_lookup_definition_r          (NxpNameSpace   *self,
							    const char     *name,
							    gboolean       rec);
NxpDefinition* nxp_name_space_lookup_definition_extended_r (NxpNameSpace   *self,
							    const char     *name,
							    NxpNameSpace  **actual_name_space,
							    gboolean        recursive);

/* --- Listeners --- */
void     nxp_name_space_add_listener     (NxpNameSpace      *self,
					  NxpNSListenerFunc  func,
					  gpointer           data,
					  NxpEvalContext    *context);
gboolean nxp_name_space_wakeup_listeners (NxpEvalContext    *context);



#endif
