/* -*- Mode: C; c-file-style: "bsd"; -*- */
#ifndef __OP_BINDINDS_H__
#define __OP_BINDINDS_H__


typedef enum {
    NXP_OP_BINDINGS_ERROR_ALREADY_BOUND = 1,
    NXP_OP_BINDINGS_ERROR_NOT_BOUND 
} NxpOpBindingsError;

/* Naming convention for operators: 

   - operators are named after their composing characters, *not*
     considering their usual function;

   - Enum name for binary operator should be NXP_OP_<CHAR1>[_<CHAR2>]
     where <CHARn> is the name of the nth character composing the
     operator. Character names can be abbreviated, but should be
     consistent (i.e. don't call = EQ in one place and EQUAL in
     another)

   - For unary operators, name should be NXP_OP_<CHAR1>[_<CHAR2>]_(LEFT|RIGHT);
     LEFT is for unary operators that are placed to the left of the operand;
     RIGHT is for unary operators that are placed to the right of the operand;

   - If the same char is repeated consectively, a number can be used
     in front of the character to indicate repetition. Example: ++
     becomes PLUS2.

*/


typedef enum {			/* ex:	(usual significance):	*/
    NXP_OP_PLUS = 1,		/* a+b	(sum)			*/
    NXP_OP_MINUS,		/* a-b	(difference)		*/
    NXP_OP_MINUS_LEFT,		/* -1	(unary minus)		*/
    NXP_OP_STAR,		/* 2*3	(multiplication)	*/
    NXP_OP_SLASH,		/* 2/3	(division)		*/
    NXP_OP_HAT,			/* e^x	(power)			*/
    NXP_OP_DOT3,		/* 1..2 (range)			*/
    NXP_OP_PLUS2_LEFT,		/* ++x	(pre-increment)		*/
    NXP_OP_MINUS2_LEFT,		/* --x	(pre-decrement)		*/
    NXP_OP_PLUS2_RIGHT,		/* x++	(post-increment)	*/
    NXP_OP_MINUS2_RIGHT,	/* x--	(post-decrement)	*/
    NXP_OP_GT,			/* a>b	(greater than)		*/
    NXP_OP_GT_EQ,		/* a>=b	(greater or equal than)	*/
    NXP_OP_LT,			/* a<b	(less than)		*/
    NXP_OP_LT_EQ,		/* a<=b	(less or equal than)	*/
    NXP_OP_EQ2,  	        /* a==b	(compares equal)	*/
    NXP_OP_EXCL_EQ,		/* a!=b	(not equal)		*/
    NXP_OP_SLASH_BSLASH,	/* a/\b	(and)			*/
    NXP_OP_BSLASH_SLASH,	/* a\/b	(or)			*/
    NXP_OP_TILDE_LEFT,		/* ~a	(not)			*/
    NXP_OP_EXCL_RIGHT,		/* a!	(factorial)		*/
    NXP_OP_EQ,  	        /* a=b	(atrib)			*/
    NXP_OP_COLON_EQ,		/* a:=b	(store)			*/
    NXP_OP_PERCENT,             /* a%b  (division remain)       */
} NxpOp;

/* number of operators */
#define NXP_NUM_OPS NXP_OP_PERCENT/* always set this to the last element in the enum! */

#include "element.h"

typedef NxpElement* (*NxpOpEvalFunc)(NxpOp op, NxpElement *left, NxpElement *right,
				     NxpEvalContext *context, GError **error);
typedef NxpElement* (*NxpOpSimpFunc)(NxpOp op, NxpElement *left, NxpElement *right,
				     NxpEvalContext *context, GError **error);


/* Please notice the difference between NxpOpBindFlags and
   NxpOpBindingFlags. The former is used as argument to
   nxp_bind_operator_function(), while the latter is a member of the
   NxpOpBinding object. */

typedef enum {
    NXP_OP_BIND_SYMMETRIC = 1 << 0 /* means order of operands is irrelevant */
} NxpOpBindFlags;


typedef enum {
    NXP_OP_BINDING_REVERSE = 1 << 0 /* means reverse order of elements
                                       in function calls */
} NxpOpBindingFlags;

typedef struct {
    guint ref_count;
    NxpOpEvalFunc eval_func;
    NxpOpSimpFunc simp_func;
    NxpOpBindingFlags flags;
} NxpOpBinding;

typedef struct _NxpOpInfo NxpOpInfo;

typedef enum {
    NXP_OP_BINARY,
    NXP_OP_UNARY_LEFT,
    NXP_OP_UNARY_RIGHT
} NxpOpKind;

struct _NxpOpInfo {
    NxpOp op;
    const char *token; /* token string as recognized by the parser */
    NxpOpKind kind;
    gint precedence;   /* the lower the number, the higher the precence */
};


void             nxp_op_bindings_init       (void);
void             nxp_op_bindings_fini       (void);
void             nxp_bind_operator_function (NxpOp            op,
					     NxpElementType   type_left,
					     NxpElementType   type_right,
					     NxpOpEvalFunc    eval,
					     NxpOpSimpFunc    simp,
					     NxpOpBindFlags   flags,
					     GError         **error);
NxpOpBinding*    nxp_get_operator_binding   (NxpOp            op,
					     NxpElementType   type_left,
					     NxpElementType   type_right,
					     GError         **error);
NxpOpBinding*    nxp_get_higher_operator_binding   (NxpOp            op,
					     NxpElement	     *left,
					     NxpElement	     *right,
					     GError         **error);
NxpOpBinding*    nxp_op_binding_new         (void);
void             nxp_op_binding_ref         (NxpOpBinding    *self);
void             nxp_op_binding_unref       (NxpOpBinding    *self);
const NxpOpInfo* nxp_op_get_info            (NxpOp            op);
void nxp_op_binding_propagate(NxpElementTypeInfo *typeinfo);




#endif /* __OP_BINDINDS_H__ */
